#include "mgP.h"
#include "mgx11P.h"
#include "windowP.h"
#include "mgx11windows.h"

#define THEIGHT 80
#define TWIDTH 40

static mgx11_sort *mgx11sort = NULL;	/* global sorting structure */
static Display *mgx11display = NULL;	/* global X display pointer */
static unsigned long mgx11colors[6][6][6];

void Xmg_setx11display(Display *dpy)
{
  XColor	col;
  Colormap	cmap;
  int		rcol, gcol, bcol, colfail;

  _mgx11c->mgx11display = dpy;

  if (mgx11display == dpy)
    return;

  mgx11display = dpy;

  if (_mgx11c->pix)
  {
    cmap = XCreateColormap(mgx11display, DefaultRootWindow(mgx11display),
			    DefaultVisual(mgx11display,
			    DefaultScreen(mgx11display)),
			    AllocNone);
  }
  else
    cmap = DefaultColormapOfScreen(DefaultScreenOfDisplay(mgx11display));

  colfail = 0;

    for (rcol=0; rcol<6; rcol++)
      for (gcol=0; gcol<6; gcol++)
        for (bcol=0; bcol<6; bcol++) {
          col.red = (unsigned short)(rcol*13107);
          col.green = (unsigned short)(gcol*13107);
          col.blue = (unsigned short)(bcol*13107);
          if (XAllocColor(mgx11display, cmap, &col))
            mgx11colors[rcol][gcol][bcol] = col.pixel;
          else
	  {
	    mgx11colors[rcol][gcol][bcol] = -1;
	    colfail = 1;
	  }
        }

}

int Xmg_initx11device()
{

  if (!mgx11sort)
  {
    mgx11sort = (mgx11_sort *)malloc(sizeof(mgx11_sort));
    mgx11sort->primnum = 1000;
    VVINIT(mgx11sort->primsort, int, mgx11sort->primnum);
    vvneeds(&(mgx11sort->primsort), mgx11sort->primnum);
    VVINIT(mgx11sort->prims, mgx11prim, mgx11sort->primnum);
    vvneeds(&(mgx11sort->prims), mgx11sort->primnum);

    mgx11sort->pvertnum = 2024;
    VVINIT(mgx11sort->pverts, Point3, mgx11sort->pvertnum);
    vvneeds(&(mgx11sort->pverts), mgx11sort->pvertnum);
/*    fprintf(stderr,"Allocated prims and vertices\n"); */

  }
  _mgx11c->mysort = mgx11sort;
  _mgx11c->myxwin = NULL;

  return 1;

}

void Xmg_initzrange()
{
/*  fprintf(stderr,"X11: initialize Z-buffer (NOT!)\n"); */
}

int Xmg_openwin(char *id)
{
  mgx11win *current;
  int      border_width = 0;
  int	   border_color, bg_color;

  if (!mgx11display)
    Xmg_setx11display(XOpenDisplay(NULL));

  border_color = WhitePixel(_mgx11c->mgx11display,
		DefaultScreen(_mgx11c->mgx11display));
  bg_color = BlackPixel(_mgx11c->mgx11display,
		DefaultScreen(_mgx11c->mgx11display));

/*  fprintf(stderr,"X11: open a window, name = %s\n",id); */

  if (_mgx11c->myxwin == NULL)
    _mgx11c->myxwin = (mgx11win *)malloc(sizeof(mgx11win));

  current = _mgx11c->myxwin;
  current->window = XCreateSimpleWindow(_mgx11c->mgx11display,
 			DefaultRootWindow(_mgx11c->mgx11display), 0, 0, TWIDTH,
			THEIGHT, border_width, border_color, bg_color);
  XStoreName(_mgx11c->mgx11display, current->window, id);
  
  current->xswa.colormap = DefaultColormap(_mgx11c->mgx11display,
			DefaultScreen(_mgx11c->mgx11display));
  current->xswa.backing_store = Always;
  current->xswa.bit_gravity = CenterGravity;
  XChangeWindowAttributes(_mgx11c->mgx11display, current->window, (CWColormap|
			CWBitGravity), &(*current).xswa);
  current->gc = DefaultGC(_mgx11c->mgx11display,
	DefaultScreen(_mgx11c->mgx11display));
  XSetForeground(_mgx11c->mgx11display, current->gc,
	WhitePixel(_mgx11c->mgx11display,
	DefaultScreen(_mgx11c->mgx11display)));
  XSetBackground(_mgx11c->mgx11display, current->gc, bg_color);
  XMapWindow(_mgx11c->mgx11display, current->window);
  XClearWindow(_mgx11c->mgx11display, current->window);
  current->mypix = XCreatePixmap(_mgx11c->mgx11display, current->window,
			TWIDTH, THEIGHT, DefaultDepth(_mgx11c->mgx11display,
				  DefaultScreen(_mgx11c->mgx11display)));
	
/*  fprintf(stderr,"bitplanes = %d\n", (int)DefaultDepth(_mgx11c->mgx11display,
				  DefaultScreen(_mgx11c->mgx11display))); */
  _mgx11c->myxwin = current;
  return 1;
}

void Xmg_setparent(Window win)
{
  _mgx11c->myxwin->parent = win;
}

void Xmg_setwin(Window win)
{
  mgx11win *current;
  int toss;
  int	   bg_color = BlackPixel(_mgx11c->mgx11display,
		DefaultScreen(_mgx11c->mgx11display));

  if (_mgx11c->myxwin == NULL)
    _mgx11c->myxwin = (mgx11win *)malloc(sizeof(mgx11win));

  _mgx11c->visible = 1;
  _mgx11c->myxwin->window = win;
  current = _mgx11c->myxwin;
  if (!_mgx11c->pix)
  {
    current->xswa.colormap = DefaultColormap(_mgx11c->mgx11display,
                        DefaultScreen(_mgx11c->mgx11display));
    current->xswa.backing_store = Always;
    current->xswa.bit_gravity = CenterGravity;
    XChangeWindowAttributes(_mgx11c->mgx11display, current->window, (CWColormap|
                        CWBitGravity), &(*current).xswa);
  }

  current->gc = DefaultGC(_mgx11c->mgx11display,
	DefaultScreen(_mgx11c->mgx11display));
  current->mypix = NULL;
  Xmg_getwinsize(&toss, &toss, &toss, &toss);
  XSetForeground(_mgx11c->mgx11display, current->gc,
	WhitePixel(_mgx11c->mgx11display,
	DefaultScreen(_mgx11c->mgx11display)));
  XSetBackground(_mgx11c->mgx11display, current->gc, bg_color);
  if (!_mgx11c->pix)
    XClearWindow(_mgx11c->mgx11display, current->window);
}

void Xmg_poswin(int x1, int y1, int x2, int y2)
{
/*  fprintf(stderr,"X11: position window %d %d %d %d\n",x1,y1,x2,y2); */
}

void Xmg_prefposwin(int x1, int y1, int x2, int y2)
{
/*  fprintf(stderr,"X11: preferred position window %d %d %d %d\n",x1,y1,x2,y2);
 */
}

void Xmg_sizewin(int x, int y)
{
  if (_mgx11c->sizelock)
    return;
/*  fprintf(stderr,"X11: size window %d %d\n",x, y); */

  if (!_mgx11c->pix)
    XResizeWindow(_mgx11c->mgx11display, _mgx11c->myxwin->window,
	(unsigned int) x, (unsigned int) y);
}

void Xmg_minsize(int x, int y)
{
/*  fprintf(stderr,"X11: minimum size of window %d %d\n",x, y); */
}

void Xmg_winconstrain()
{
/*  fprintf(stderr,"X11: window constrain\n"); */
}

void Xmg_titlewin(char *name)
{
/*  fprintf(stderr,"X11: title window: %s\n",name); */
}

void Xmg_closewin(mgx11win *this)
{
/*  fprintf(stderr,"X11: close a window\n"); */
}

void Xmg_choosewin(int id)
{
/*  fprintf(stderr,"X11: choose a window\n"); */
}

void Xmg_flush()
{
  XFlush(_mgx11c->mgx11display);
/*  fprintf(stderr,"X11: XFlush\n"); */
}

void Xmg_newdisplaylist()
{
/*  fprintf(stderr,"X11: Start Display List\n"); */

  _mgx11c->mysort->cprim = 0;
  _mgx11c->mysort->cvert = 0;
  _mgx11c->mysort->maxverts = 0;
  _mgx11c->znudgeby = 0.0;

}

void Xmg_add(int primtype, int ndata, void *data)
{
  HPoint3 *vt = (HPoint3 *)data;
  float *col = (float *)data;
  Point3 *vts;
  int i;
  static unsigned long keepcolor, keepecolor;
  static mgx11prim *prim;
  static int *primp;
  static int numverts;
  static float average_depth;
  
  switch (primtype)
  {
	case MGX_BGNLINE:
	  average_depth = 0.0;
          prim = &(VVEC(_mgx11c->mysort->prims, mgx11prim)[_mgx11c->mysort->cprim]);
          primp = &(VVEC(_mgx11c->mysort->primsort, int)[_mgx11c->mysort->cprim]);
	  prim->mykind = PRIM_LINE;
	  prim->index = _mgx11c->mysort->cvert;
	  prim->depth = -100000; /* very far behind the viewer */
	  numverts = 0;

          VVEC(_mgx11c->mysort->primsort, int)[_mgx11c->mysort->cprim] =
		_mgx11c->mysort->cprim;

	  if (!(_mgc->has & HAS_S2O))
	    mg_findS2O();
	  mg_findO2S();
	  break;

	case MGX_BGNPOLY:
	  average_depth = 0.0;
	  /*fprintf(stderr,"Begin drawing polygon\n");*/
          prim = &(VVEC(_mgx11c->mysort->prims, mgx11prim)[_mgx11c->mysort->cprim]);
	  prim->mykind = PRIM_POLYGON;
	  prim->index = _mgx11c->mysort->cvert;
	  prim->depth = -100000; /* very far behind the viewer */
	  numverts = 0;
	  
          VVEC(_mgx11c->mysort->primsort, int)[_mgx11c->mysort->cprim] =
		_mgx11c->mysort->cprim;

	  if (!(_mgc->has & HAS_S2O))
	    mg_findS2O();
	  mg_findO2S();
	  break;

	case MGX_BGNEPOLY:
	  average_depth = 0.0;
          prim = &(VVEC(_mgx11c->mysort->prims, mgx11prim)[_mgx11c->mysort->cprim]);
	  prim->mykind = PRIM_EPOLYGON;
	  prim->index = _mgx11c->mysort->cvert;
	  prim->depth = -100000; /* very far behind the viewer */
	  numverts = 0;
	  
          VVEC(_mgx11c->mysort->primsort, int)[_mgx11c->mysort->cprim] =
		_mgx11c->mysort->cprim;

	  if (!(_mgc->has & HAS_S2O))
	    mg_findS2O();
	  mg_findO2S();
	  break;

	case MGX_VERTEX:
	  for (i=0; i<ndata; i++)
	  {
            vts = &(VVEC(_mgx11c->mysort->pverts, Point3)[_mgx11c->mysort->cvert]);
	    if (HPt3TransPt3(_mgc->O2S, &(vt[i]), vts) < 0)
	      prim->mykind = PRIM_INVIS; /* partially behind clipping
						plane */

	    vts->z += _mgx11c->znudgeby;
	
/*	    fprintf(stderr,"vertex %f %f %f - %f %f %f\n",vt[i].x, vt[i].y,
		vt[i].z, vts->x, vts->y, vts->z);*/
	

	    _mgx11c->mysort->cvert++;
	    numverts++;
	    if (_mgx11c->mysort->cvert > _mgx11c->mysort->pvertnum)
	    {
	       _mgx11c->mysort->pvertnum*=2;
	       vvneeds(&(_mgx11c->mysort->pverts), _mgx11c->mysort->pvertnum);
	    }

	    if (vts->z > prim->depth)
	    {
	      prim->depth = vts->z;
	    }

	  average_depth += vts->z;

	  }
	  break;

	case MGX_COLOR:
	  keepcolor = mgx11_nearestRGB(col[0],col[1],col[2]);
	  if (keepcolor == -1)
	    keepcolor = WhitePixel(_mgx11c->mgx11display,
			DefaultScreen(_mgx11c->mgx11display));
	  break;

	case MGX_ECOLOR:
	  keepecolor = mgx11_nearestRGB(col[0],col[1],col[2]);
	  if (keepecolor == -1)
	    keepecolor = BlackPixel(_mgx11c->mgx11display,
			DefaultScreen(_mgx11c->mgx11display));
	  break;

	case MGX_END:
	  prim->numvts = numverts;
	  if (numverts > _mgx11c->mysort->maxverts)
	    _mgx11c->mysort->maxverts = numverts;
          average_depth += prim->depth;
          average_depth /= (float)(numverts+1);
	  prim->depth = average_depth;
	  prim->pixel = keepcolor;
	  prim->epixel = keepecolor;
	  _mgx11c->mysort->cprim++;
	  if (_mgx11c->mysort->cprim > _mgx11c->mysort->primnum)
	  {
	    _mgx11c->mysort->primnum*=2;
	    vvneeds(&(_mgx11c->mysort->prims), _mgx11c->mysort->primnum);
	    vvneeds(&(_mgx11c->mysort->primsort), _mgx11c->mysort->primnum);
	  }
	  break;

	default:
	  fprintf(stderr,"unknown type of primitive.\n");
	  break;
  }
}

int Xmg_primcomp(const void *a, const void *b)
{
  mgx11prim *prim =
	VVEC(_mgx11c->mysort->prims, mgx11prim);

/*  fprintf(stderr,"sort:%f %f\n", (*(mgx11prim **)a)->depth,
		(*(mgx11prim **)b)->depth);*/

  if (prim[*(int*)a].depth < prim[*(int*)b].depth)
    return 1;
  else
    return -1;

}

void Xmg_sortdisplaylist()
{
  static int *primp;
  primp = VVEC(_mgx11c->mysort->primsort, int);
  qsort(primp, _mgx11c->mysort->cprim, sizeof(int), &Xmg_primcomp);
/*  fprintf(stderr,"X11: Sort Display List\n"); */
}

void Xmg_showdisplaylist()
{
  Point3 *vts, *vt;
  int ref, i, vloc, *primp;
  mgx11prim *prim, *prim2;
  Display *dpy = _mgx11c->mgx11display;
  mgx11win *curr = _mgx11c->myxwin;
  Drawable win = _mgx11c->myxwin->window;
  Drawable pix = _mgx11c->myxwin->mypix;
  GC gc = _mgx11c->myxwin->gc;
  XPoint *xvts = (XPoint *) malloc(sizeof(
			XPoint)*(_mgx11c->mysort->maxverts+1));
  static int width;
  static int height;
  static unsigned long bgcolor;

  WnGet(_mgc->win, WN_XSIZE, &width);
  WnGet(_mgc->win, WN_YSIZE, &height);
/*  fprintf(stderr,"size is %dx%d\n",width, height); */

/*  bgcolor =  mgx11colors[(int)(_mgc->background.r*5.1)]
			[(int)(_mgc->background.g*5.1)]
			[(int)(_mgc->background.b*5.1l)];*/

  bgcolor = mgx11_nearestRGB(_mgc->background.r, _mgc->background.g,
				_mgc->background.b);
  if (bgcolor == -1)
    XSetForeground(dpy, gc, BlackPixel(_mgx11c->mgx11display,
			DefaultScreen(_mgx11c->mgx11display)));
  else
    XSetForeground(dpy, gc, bgcolor);

  XFillRectangle(dpy, pix, gc, 0, 0, (unsigned int)width,
		(unsigned int)height);
/*  fprintf(stderr,"X11: Show Display List\n"); */

  primp = VVEC(_mgx11c->mysort->primsort, int);
  prim2 = VVEC(_mgx11c->mysort->prims, mgx11prim);
  vts = VVEC(_mgx11c->mysort->pverts, Point3);

/*  fprintf(stderr,"number of primitives %d\n",_mgx11c->mysort->cprim); */
  for (ref = 0; ref < _mgx11c->mysort->cprim; ref++)
  {
    prim = &(prim2[primp[ref]]);
    switch (prim->mykind)
    {
      case PRIM_POLYGON:
        vloc = prim->index;
        for (i = 0; i< prim->numvts; i++)
        {
          xvts[i].x = (short)vts[vloc].x;
          xvts[i].y = (short)vts[vloc].y;
          vloc++;
        }

        XSetForeground(dpy, gc, prim->pixel);
        XFillPolygon(dpy, pix, gc, xvts, prim->numvts, Convex, CoordModeOrigin);
        break;

      case PRIM_EPOLYGON:
        vloc = prim->index;
        for (i = 0; i< prim->numvts; i++)
        {
          xvts[i].x = (short)vts[vloc].x;
          xvts[i].y = (short)vts[vloc].y;
          vloc++;
        }
        xvts[i].x = (short)vts[prim->index].x; /* add extra point */
        xvts[i].y = (short)vts[prim->index].y; /* for closed polyline */

        XSetForeground(dpy, gc, prim->pixel);
        XFillPolygon(dpy, pix, gc, xvts, prim->numvts, Convex, CoordModeOrigin);
        XSetForeground(dpy, gc, prim->epixel);
        XDrawLines(dpy, pix, gc, xvts, prim->numvts+1, CoordModeOrigin);
        break;

      case PRIM_LINE:
        vloc = prim->index;
        for (i = 0; i< prim->numvts; i++)
        {
          xvts[i].x = (short)vts[vloc].x;
          xvts[i].y = (short)vts[vloc].y;
          vloc++;
        }
        XSetForeground(dpy, gc, prim->epixel);
        XDrawLines(dpy, pix, gc, xvts, prim->numvts, CoordModeOrigin);
        break;

      case PRIM_INVIS:
	break;

      default:
        break;
    }
  }
  XCopyArea(dpy, pix, win, gc, 0, 0, (unsigned int)width,
		(unsigned int)height, 0,0);
  free(xvts);
  Xmg_flush();
}

void Xmg_getwinsize(int *xsize, int *ysize, int *xorig, int *yorig)
{
  Display *dpy = _mgx11c->mgx11display;
  Window dpyroot;
  Window Toss;
  Drawable win = _mgx11c->myxwin->window;
  unsigned int width, height, border_width, depth;
  int xpos, ypos, xold, yold;

/*  fprintf(stderr,"X11: Get Window Size\n"); */
  if (_mgx11c->visible)
  {
    XGetGeometry(dpy, win, &dpyroot, &xpos, &ypos, &width,
		&height, &border_width, &depth);
    *xsize = width;
    *ysize = height;

    if (_mgx11c->pix)
    {
      *xorig = 0; *yorig = 0;
    }
    else
    if (XTranslateCoordinates(dpy, win, dpyroot, 0, height-1, &xpos, &ypos,
					&Toss))
    {
      *xorig=xpos; *yorig=HeightOfScreen(DefaultScreenOfDisplay(dpy)) - ypos;
    }
    else
      {
        *xorig = 0; *yorig = 0;
      }

  }
  else
  {
    *xsize = 0; *ysize = 0;
  }
  WnGet(_mgc->win, WN_XSIZE, &xold);
  WnGet(_mgc->win, WN_YSIZE, &yold);
  if ((xold != width) || (yold != height) || (_mgx11c->myxwin->mypix == NULL))
  {
    if (_mgx11c->myxwin->mypix != NULL)
      XFreePixmap(dpy, _mgx11c->myxwin->mypix);
/*    fprintf(stderr,"create pixmap %d x %d\n",width,height); */
    _mgx11c->myxwin->mypix = XCreatePixmap(dpy, win, width, height,
				DefaultDepth(dpy, DefaultScreen(dpy)));
  }
}

unsigned long mgx11_nearestRGB(float r, float g, float b)
{
  return mgx11colors[(int)(r*5.0+.5)][(int)(g*5.0+.5)] [(int)(b*5.0+.5)];
}
